/*
   Extended translation utility, written by Wojciech M. Zabolotny
   wzab@ise.pw.edu.pl
   This source is published as "public domain"
   This file implements the command line scanning and pattern building
   for the xtr utility
*/
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <time.h>
#include <getopt.h>
#include "xtr.h"
//You can change the value below. The program is buffer overflow safe,
//but will work unreliably if too long line is found
#define MAX_LINE_LEN (1000)

void DefStrScan(char * defstr)
{
  Pattern * PatS, *PatR, *CurPat;
  PatS = (Pattern *) malloc(sizeof(Pattern));
  if(!PatS) exit(ERR_MEMORY);
  PatInit(PatS);
  PatR = (Pattern *) malloc(sizeof(Pattern));
  if(!PatS) exit(ERR_MEMORY);
  PatInit(PatR);
  /*
     Patterns are ready, now scan the definition string
  */
  CurPat=PatS;
  while(*defstr)
    {
      char c=*(defstr++);
      switch(c)
	{
	case '\\':
	  /*
	    If these is the backslash character, it requires special
	    handling
	    First we check the next character
	  */
	  {
	    char c1=*(defstr++);
	    switch(c1)
	      {
		char c3;
	      case 'o':
		/*
		  We read the three digit octal number
		*/
		{
		  int onum=0;
		  int i;
		  for (i=0;i<3;i++)
		    {
		      c3=*(defstr++);
		      if((c3<'0') || (c3>'7'))
			{
			  fprintf(stderr,"There must be EXACTLY three octal digits"
				  "after the '\\o' sequence in: %s\n",optarg);
			  exit(ERR_SYNTAX);
			}
		      else
			{
			  onum=onum*8+(c3-'0');
			}
		    }
		  if(onum>255)
		    {
		      fprintf(stderr,"The character's code must be not greater"
			      "than 255 in: %s\n",optarg);
		      exit(ERR_SYNTAX);
		    }
		  else
		    {
		      PatAddChar(CurPat,(char) onum);
		    }
		}
		break;
	      case 'd':
		/*
		  We read the three digit decimal number
		*/
		{
		  int onum=0;
		  int i;
		  for (i=0;i<3;i++)
		    {
		      c3=*(defstr++);
		      if((c3<'0') || (c3>'9'))
			{
			  fprintf(stderr,"There must be EXACTLY three decimal digits"
				  "after the '\\d' sequence in: %s\n",optarg);
			  exit(ERR_SYNTAX);
			}
		      else
			{
			  onum=onum*10+(c3-'0');
			}
		    }
		  if(onum>255)
		    {
		      fprintf(stderr,"The character's code must be not greater"
			      "than 255 in: %s\n",optarg);
		      exit(ERR_SYNTAX);
		    }
		  else
		    {
		      PatAddChar(CurPat,(char) onum);
		    }
		}
		break;
	      case 'x':
		/*
		  We read the two digit hexadecimal number
		*/
		{
		  int onum=0;
		  int i;
		  for (i=0;i<2;i++)
		    {
		      c3=*(defstr++);
		      if((c3>='0') && (c3<='9'))
			{
			  onum=onum*16+(c3-'0');
			}
		      else if((c3>='a') && (c3<='f'))
			{
			  onum=onum*16+(c3-'a'+10);
			}
		      else if((c3>='A') && (c3<='F'))
			{
			  onum=onum*16+(c3-'A'+10);
			}
		      else
			{
			  fprintf(stderr,"There must be EXACTLY two hexadecimal digits"
				  "after the '\\x' sequence in: %s\n",optarg);
			  exit(ERR_SYNTAX);
			}
		    }
		  if(onum>255)
		    {
		      fprintf(stderr,"The character's code must be not greater"
			      "than 255 in: %s\n",optarg);
		      exit(ERR_SYNTAX);
		    }
		  else
		    {
		      PatAddChar(CurPat,(char) onum);
		    }
		} /* case 'x': */
		break;
	      default:
		PatAddChar(CurPat,c1);
		break;
	      } /* switch (c1) */
	  } /* case '//' */
	  break;
	case ',':
	  /*
	    If this is comma, we have to proceed with next pattern
	  */
	  if(CurPat==PatS)
	    {
	      CurPat=PatR;
	    }
	  else
	    {
	      /*
		This is an error, caused probably by the presence of
		two commas in the definition string
	      */
	      fprintf(stderr,"There are two commas in the definition"
		      "string: %s",optarg);
	      exit(ERR_SYNTAX);
	    }
	  break;
	default:
	  PatAddChar(CurPat,c);
	} /* switch (c) */
    } /* while (*(defstr++)) */
  /*
    The patterns are filled in, so we can add them
    to the pattern list
  */
  PatListAddSet(PatS,PatR);
}

void CmdLineScan(int argc,char * const argv[])
{
  while(1)
    {
      int opt=getopt(argc,argv,"t:f:");
      if (opt==-1) break;
      switch(opt)
	{
	case 't':
	  {
	    /*
	      Option -t should be followed by the string containing
	      two substrings separated by comma. First substring
	      describes the searched pattern, the second one -
	      its substitute.
	      The patterns may contain: common characters,
	      if comma should be present in pattern, then it should
	      be preceded by backslash.
	      The codes of characters entered in one of three ways:
	      \x?? - The hexadecimal code (there must be exactly two
	      digits)
	      \o??? - The octal code (there must be exactly three
	      digits)
	      \d??? - The decimal code (there must be exactly three
	      digits)
	    */
	    DefStrScan(optarg);
	  } /* case 't': */
	  break;
	case 'f':
	  /*
	    Read patterns definition from the file
	  */
	  {
	    FILE * fdef;
	    char linebuf[MAX_LINE_LEN];
	    fdef=fopen(optarg,"rt");
	    if(!fdef)
	      {
		fprintf(stderr,"Can not read pattern definitions"
			" from file: %s\n",optarg);
		exit(ERR_SYNTAX);
	      }
	    while(fgets(linebuf,MAX_LINE_LEN,fdef))
	      {
		/*
		  Cut the line before the EOL character
		*/
		char * c=linebuf;
		while(*c)
		  {
		    if (*c=='\n')
		      {
			*c=0;
			break;
		      }
		    c++;
		  }
		DefStrScan(linebuf);
	      }
	  }
	  break;
	} /* switch (opt) */
    } /* while (1) */
}
