/*
  Extended translation utility written by Wojciech M. Zabolotny
  wzab@ise.pw.edu.pl
  This source is published as "public domain"
*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "xtr.h"
#define PatStep (10)
#define min(x,y) ((x)<(y)) ? (x) : (y)
#define INBUFSIZE (256)
#define OUTBUFSIZE (256)
FILE * strin;
FILE * strout;
char InBuf[INBUFSIZE];
char OutBuf[OUTBUFSIZE];
void CmdLineScan(int argc,char * const argv[]);
/*
   The program originally developed in "C++" has been rewritten in "C"
   for better portability
*/

/* Functions operating on Pattern structures */

/* Pattern structure initilization ("constructor") */
void PatInit(Pattern * p)
{
  p->len = 0;
  p->data = NULL;
  p->size = 0;
}

/* Adding of one character to Pattern structure */
void PatAddChar(Pattern * p, char c)
{
  if ( p->len >= p->size )
    {
      /* There is no place for new character - reallocate the buffer */
      int newsize=p->size+PatStep;
      p->data = (char *) realloc(p->data,newsize);
      if(! p->data ) exit( ERR_MEMORY ); /* Memory allocation error */
      p->size = newsize;
    }
  /* Now we have enough place for the next character */
  p->data[p->len++]=c;
}

/* Cleaning up of Pattern structure ("destructor") */
void PatClean(Pattern * p)
{
  free(p->data);
  p->data = NULL;
  p->len = 0;
  p->size = 0;
  free(p);
}

/* Function for comparison of Pattern structures
   Return values:
   -2 : *p1 < *p2
   -1 : *p2 contains at the beginning contents of *p1
   0 : *p1 == *p2
   +1 : *p1 contains at the beginning contents of *p2
   +2 : *p1 > *p2
*/
int PatCmp(Pattern * p1, Pattern * p2)
{
  int len,i;
  /* First we check which pattern is longer */
  len = min (p1->len, p2->len);
  for (i=0;i<len;i++)
    {
      if (p1->data[i] < p2->data[i]) return -2;
      if (p1->data[i] > p2->data[i]) return 2;
    }
  /* One string is contained at the begining of the another one
     We are comparing lenghts
  */
  if (p1->len < p2->len) return -1;
  if (p1->len > p2->len) return 1;
  return 0;
}

/* Structure holding set of patterns (one for searching, and the another
   one for replacing)
*/
typedef struct
{
  Pattern * srch; /* Pattern to be replaced */
  Pattern * repl; /* Replacement pattern */
}  PatSet;

/* Variables implementing the patterns' sets list */
int PatListLen=0; /* Number of patterns' sets */
int PatListSize=0; /* Lenght of list of patterns' sets */
PatSet * PatList=NULL; /* Patterns' sets list */

/* Buffer for storing of pattern read from the input */
Pattern PBuf;

/* Functions dealing with patterns' sets list */

/* Cleaning of patterns' sets list */
void PatListClean()
{
  int i;
  for (i=0;i<PatListLen;i++)
    {
      PatClean(PatList[i].srch);
      PatClean(PatList[i].repl);
      free(PatList);
      PatListLen=0;
      PatListSize=0;
      PatList=NULL;
    }

}
/* Adding of new patterns set */
void PatListAddSet(Pattern * PatSearch, Pattern * PatRepl)
{
  if ( PatListLen >= PatListSize )
    {
      /* There is no place for the new patterns set - reallocate the 
buffer */
      int newsize=PatListSize+PatStep;
      PatList = (PatSet *) realloc(PatList,newsize * sizeof(PatSet));
      if(! PatList ) exit( ERR_MEMORY ); /* Memory allocation error */
      PatListSize = newsize;
    }
  /* Now we have enough place for the next set */
  PatList[PatListLen].srch = PatSearch;
  PatList[PatListLen++].repl = PatRepl;
}

/* Function for comparison of sets of patterns (used for sorting) */
int PatSetCmp(const void * p1, const void * p2)
{
  PatSet * ps1;
  PatSet * ps2;
  ps1=(PatSet *) p1;
  ps2=(PatSet *) p2;
  return PatCmp(ps1->srch, ps2->srch);
}

/* Function for sorting of sets of patterns */
void PatListSort()
{
  qsort(PatList,PatListLen,sizeof(PatSet),PatSetCmp);
}
#ifdef DEBUG
/* Function for printing of sets of patterns (used for debugging only) */
void PatSetPrint(PatSet * p)
{
  Pattern * pr;
  int i;
  pr=p->srch;
  for (i=0;i<pr->len;i++)
    {
      fputc(pr->data[i],stderr);
    }
  fprintf(stderr,"->");
  pr=p->repl;
  for (i=0;i<pr->len;i++)
    {
      fputc(pr->data[i],stderr);
    }
  fputc('\n',stderr);
}

/* Function for building a new patterns set from two "C" strings */
void PSB(char * ps, char * pr)
{
  Pattern * PatS, * PatR;
  int i;
  PatS=(Pattern *) malloc(sizeof(Pattern));
  if(!PatS) exit(ERR_MEMORY);
  PatInit(PatS);
  for (i=0;i<strlen(ps);i++) PatAddChar(PatS,ps[i]);
  PatR=(Pattern *) malloc(sizeof(Pattern));
  if(!PatR) exit(ERR_MEMORY);
  PatInit(PatR);
  for (i=0;i<strlen(pr);i++) PatAddChar(PatR,pr[i]);
  PatListAddSet(PatS,PatR);
}
#endif /* DEBUG */

/* The CheckBuf functions checks if PBuf buffer matches any pattern
   from the PatList
   Returned values:
   -2 - There exists at least one pattern matching, but longer than PBuf
   -1 - There is no matching pattern at all
   >=0 - Number of the longest matching pattern (may be shorter then PBuf)

   When we recheck the buffer after a new character is added, there is 
no need
   to check all the patterns once again. The matching patterns area may 
be only narrowed
   So we could use two variables: LLimit & HLimit to delimit the search 
range.

   To improve the function, a binary search instead of a linearone 
should be
   used. Sorry, I've not found enough time to implement it
*/
int CheckBuf(int mode)
{
  int j;
  //The flags below describe state of the buffer
  //PShorter - there is a matching pattern, but it is shorter or of the 
same length as a buffer
  //LShorter - length of the shorter (or of the same length) pattern
  int PShorter=-1, LShorter=0;
  static int LLimit,HLimit;
  if( (mode==0) || (mode ==1) ) {
    /* We should initialize the function - we set the LLimit to 0, and 
HLimit to
       length of the pattern set */
    LLimit=0;
    HLimit=PatListLen;
  }
  if(mode==0) return 0; //We only perform the initialization
  if (PBuf.len==0) return -2; /* Zero lenght - it must match any longer 
pattern */
  //
  for(j=LLimit;(j<PatListLen) && (j<HLimit);j++)
    {
      int res=PatCmp(&PBuf,PatList[j].srch);
      if(res==-1) return -2;
      if(res==1 || res==0)
	{
	  if (LShorter<PatList[j].srch->len)
	    {
	      PShorter=j;
	      LShorter=PatList[j].srch->len;
	    }
	}
    }
  return PShorter;
}

//WriteBuf function - called with the number of a matching pattern
void WriteBuf(int MatchPat)
{
  if (MatchPat<0)
    {
      //Write the first character from the buffer, and shorten the buffer
      int i;
      fputc(PBuf.data[0],strout);
      if(PBuf.len >= 1)
	{
	  for(i=1;i<PBuf.len;i++) PBuf.data[i-1]=PBuf.data[i];
	  PBuf.len--;
	}
    }
  else
    {
      //There is a matching pattern, so write its translation
      //and remove the pattern from the buffer
      int i,PSize;
      PSize=PatList[MatchPat].srch->len;
      if (PBuf.len < PSize) exit(ERR_INTERNAL);
      for(i=PSize; i<PBuf.len; i++) PBuf.data[i-PSize]=PBuf.data[i];
      PBuf.len-=PSize;
      //Now we write the replacement
      fwrite(PatList[MatchPat].repl->data,1,
	     PatList[MatchPat].repl->len,strout);
    }
}
int main(int argc, char * argv[])
{
  int i;
  int NewChar;
  //strin=fopen("test","r");
  //strout=fopen("test.tr","w");
  strin=stdin;
  strout=stdout;
  //At the begining we try to switch on the buffering of input & output 
streams
  setbuffer(strin,InBuf,INBUFSIZE);
  setbuffer(strout,OutBuf,OUTBUFSIZE);
  CmdLineScan(argc,argv);
  //Debugging: print out the unsorted pattern's set
  //fprintf(stderr,"Unsorted patterns:\n");
  //for (i=0;i<PatListLen;i++) PatSetPrint(&(PatList[i]));
  //We sort the patterns
  PatListSort();
  //Debugging: print out the sorted pattern's set
  //fprintf(stderr,"Sorted patterns:\n");
  //for (i=0;i<PatListLen;i++) PatSetPrint(&(PatList[i]));
  //Now we check if no pattern appears twice
  for (i=0;i<(PatListLen-1);i++)
    {
      if(!PatSetCmp(&(PatList[i]),&(PatList[i+1])))
	{
	  fprintf(stderr,"Searched patterns must be unique!\n");
	  exit(ERR_SYNTAX);
	}
    }
  CheckBuf(0); //We initialize the buffer checking function
  while( (NewChar=fgetc(strin)) != EOF )
    {
      int j;
      //We add a new character to the buffer
      PatAddChar(&PBuf,(char) NewChar);
      //We check if there is still the matching pattern
      j=CheckBuf(2); /* The new character is added, so we can only 
"recheck" the buffer */
      if(j != -2)
	{
	  //Remove the leading characters from the buffer, until there is a match
	  while (j != -2)
	    {
	      WriteBuf(j);
	      j=CheckBuf(1); /* Begining of the buffer has changed so we need 
to start
				matching of the patterns from the begining */
	    }
	}
    }
  //If we've arrived there - the end of input stream is arrived
  //So we write the buffer's contents until it has a zero length
  while(PBuf.len > 0)
    {
      int j;
      j=CheckBuf(1);
      WriteBuf(j);
    }
  /*
    We free the memory (it could be done by the operating system itself,
    but some OS'es are broken...)
  */
  PatListClean();
  return 0;
}

